# -*- coding: utf8 -*-

#
# El desparramador arma una matriz que es la superposición de los
# olores de todos los elementos que haya de un tipo.
# 
# Se podría decir que es el "mapa del olor".
#
# No es elegante: quedaron cuatro funciones recursivas, que aunque
# muy parecidas, no son iguales. 
#
# Pero cumple con los requisitos:
# - "desparrama" la señal
# - corta en los stoppers
# - pero los rodea levemente
#
# Si se quisiera mejorar, habría que componer cada cuadro con parte
# de los que pueden llegar hasta él (en la actual, si se llega al cuadro
# tiene el mismo valor si se llega de uno que si se llega de 3, lo cual no
# es muy "real".
#

import pygame

# es cuanto hay que tocar x,y para ir hacia ese lado
DELTAS = {0:(0, -1), 1:(1, -1), 2:(1, 0), 3:(1, 1), 4:(0, 1), 5:(-1, 1), 6:(-1, 0), 7:(-1, -1)}


class Desparramador(object):
    def __init__(self, siz_x, siz_y):
        self.siz_x = siz_x
        self.siz_y = siz_y

    def __call__(self, stoppers, potencia, origenes):
        matrices = []
        for x,y in origenes:
            m = self._armaMat(stoppers, potencia, x, y)
            matrices.append(m)

        # mergeamos las matrices
        topvalor = 0
        matriz = []
        for x in range(self.siz_x):
            fila = []
            for y in range(self.siz_y):
                val = sum(m[x][y] for m in matrices)
                topvalor = max(topvalor, val)
                fila.append(val)
            matriz.append(fila)
        return (matriz, topvalor)

    def _armaMat(self, stoppers, potencia, nx, ny):
        siz_x, siz_y = self.siz_x, self.siz_y

        # armamos la matriz base
        self.matriz = matriz = []
        for x in range(siz_x):
            fila = []
            for y in range(siz_y):
                if stoppers[x][y]:
                    fila.append(-1)
                else:
                    fila.append(0)
            matriz.append(fila)

        # propagamos
        self._propagArriba(nx, ny, potencia)
        self._propagAbajo(nx, ny, potencia)
        self._propagIzquierda(nx, ny, potencia)
        self._propagDerecha(nx, ny, potencia)

        # corregimos ese valor negativo, a cero
        for x in range(siz_x):
            for y in range(siz_y):
                if matriz[x][y] < 0:
                    matriz[x][y] = 0

        return matriz


    def _propagArriba(self, x0, y0, val):
        self.matriz[x0][y0] = val
        val /= 2
        if not val:
            return

        # subimos una linea
        yn = y0 - 1
        if yn < 0:
            return

        # explotamos tres
        for xn in range(x0-1, x0+2):
            if 0 <= xn < self.siz_x:
                if self.matriz[xn][yn] == 0:
                    self._propagArriba(xn, yn, val)

    def _propagAbajo(self, x0, y0, val):
        self.matriz[x0][y0] = val
        val /= 2
        if not val:
            return

        # bajamos una linea
        yn = y0 + 1
        if yn >= self.siz_y:
            return

        # explotamos tres
        for xn in range(x0-1, x0+2):
            if 0 <= xn < self.siz_x:
                if self.matriz[xn][yn] == 0:
                    self._propagAbajo(xn, yn, val)
        
    def _propagIzquierda(self, x0, y0, val):
        self.matriz[x0][y0] = val
        val /= 2
        if not val:
            return

        # una col a la izq
        xn = x0 - 1
        if xn < 0:
            return

        # explotamos tres
        for yn in range(y0-1, y0+2):
            if 0 <= yn < self.siz_y:
                if self.matriz[xn][yn] == 0:
                    self._propagIzquierda(xn, yn, val)

    def _propagDerecha(self, x0, y0, val):
        self.matriz[x0][y0] = val
        val /= 2
        if not val:
            return

        # una col a la izq
        xn = x0 + 1
        if xn >= self.siz_x:
            return

        # explotamos tres
        for yn in range(y0-1, y0+2):
            if 0 <= yn < self.siz_y:
                if self.matriz[xn][yn] == 0:
                    self._propagDerecha(xn, yn, val)
        

class Terreno(object):
    def __init__(self, surface, arista, siz, init):
        self.surface = surface
        self.desparram = Desparramador(*siz)
        self.arista = arista
        self.siz_x, self.siz_y = siz
        self.init_x, self.init_y = init
        self.elementos = {}
        self.potenciaOlor = dict(Comida=50, Hormiguero=100, Hormiga=15)
        self.dispersion = False

        # construimos las baldosas
        baldosas = pygame.Surface((arista*self.siz_x, arista*self.siz_y))
        baldosas.fill((255, 255, 255))
        initx, inity = self.init_x, self.init_y
        for x in range(self.siz_x):
            for y in range(self.siz_y):
                pygame.draw.rect(baldosas, (230, 230, 230), (x*arista+1, y*arista+1, arista-2, arista-2))
        self.baldosas = baldosas

        # armamos las matrices
        self.matrizVacia = []
        self.stoppers = []
        for i in range(self.siz_x):
            f_stoppers = []
            f_vacia = []
            for j in range(self.siz_y):
                f_stoppers.append(False)
                f_vacia.append(0)
            self.stoppers.append(f_stoppers)
            self.matrizVacia.append(f_vacia)

    def limpiarBaldosas(self):
        self.surface.blit(self.baldosas, (self.init_x, self.init_y))
        pygame.display.update()

    def vaciaBaldosa(self, pos):
        nx,ny,x0,y0 = self._cualBaldosa(pos)
        for elm,(x,y) in self.elementos.items():
            if x == nx and y == ny:
                del self.elementos[elm]
        self.redibujar()

    def _cualBaldosa(self, pos):
        nx = (pos[0]-self.init_x) // self.arista
        ny = (pos[1]-self.init_y) // self.arista
        x0 = self.init_x + nx * self.arista
        y0 = self.init_y + ny * self.arista
        return nx, ny, x0, y0

    def numero(self, x, y, nro):
        font = pygame.font.Font(None, 20)
        sfc = font.render(str(nro), True, (255,0,0))
        w,h = font.size(str(nro))
        ar = self.arista
        dx = (ar-w) / 2
        dy = (ar-h) / 2
        self.surface.blit(sfc, (self.init_x+x*ar+dx, self.init_y+y*ar+dy))

                
    def tirarObjeto(self, elemento, pos):
        nx,ny,x0,y0 = self._cualBaldosa(pos)

        # guardamos
        obj = elemento.clase(nx, ny, elemento)
        self.elementos[obj] = (nx,ny)
        if elemento.nombre in ("Agua",):
            self.stoppers[nx][ny] = True

        # dibujamos
        self.surface.blit(elemento.sfc, (x0,y0))
        pygame.display.update()

    def moverObjeto(self, elemento, new_x, new_y):
        pisados = [elm for elm,(x,y) in self.elementos.items() if x==new_x and y==new_y]
        self.elementos[elemento] = (new_x, new_y)
        self.redibujar()
        return pisados

    def sacarObjeto(self, obj):
        del self.elementos[obj]
        self.redibujar()

    def redibujar(self):
        self.limpiarBaldosas()
        for elm,(nx,ny) in self.elementos.items():
            self.surface.blit(elm.sfc, (self.init_x+nx*self.arista, self.init_y+ny*self.arista))
        pygame.display.update()

    def hideDispersion(self):
        if self.dispersion:
            self.dispersion = False
            self.redibujar()

    def _calcDispersion(self, tipo):
        origenes = [v for k,v in self.elementos.items() if k.nombre == tipo]
        if not origenes:
            return (self.matrizVacia, 0)

        return self.desparram(self.stoppers, self.potenciaOlor[tipo], origenes)

    def showDispersion(self, tipo):
        (matriz, topvalor) = self._calcDispersion(tipo)
        self.dispersion = True

        # dibujamos las matrices
        initx, inity = self.init_x, self.init_y
        arista = self.arista
        balanceador = topvalor / 200.0
        for x in range(self.siz_x):
            for y in range(self.siz_y):
                if not balanceador:
                    valcol = 255
                else:
                    valcol = 255 - matriz[x][y] / balanceador
                color = (valcol,)*3
                pygame.draw.rect(self.surface, color, (initx+x*arista+1, inity+y*arista+1, arista-2, arista-2))

        # volvemos a poner lo que justamente estamos viendo cómo se propaga
        elementos = [x for x in self.elementos.keys() if x.nombre == tipo]
        for elm in elementos:
            nx,ny = self.elementos[elm]
            self.surface.blit(elm.sfc, (initx+nx*arista, inity+ny*arista))
        pygame.display.update()
        
    def getAlrededor(self, x, y, quieren):
#        print "getAlrededor de %d,%d buscando %s" % (x, y, quieren)
        (matriz, topvalor) = self._calcDispersion(quieren)

        valores = []
        for i in range(8):
            dx,dy = DELTAS[i]
#            print i,dx,dy
            nx,ny = x+dx, y+dy
            if not (0 <= nx < self.siz_x):
                val = None
            elif not (0 <= ny < self.siz_y):
                val = None
            elif self.stoppers[nx][ny]:
                val = None
            else:
                val = matriz[nx][y+dy]
            valores.append(val)
#        print "getAlrededor, final:", valores 
        return valores
        

